// LifePo4 Cell balancer  - zjednodusena verzia
// external reference on pin 5 Aref (2.048V)
// measured Ubat on pin 3 throught divider (Ubat/2)
//
// in sleep mode it takes about 135 uA, while 125 uA takes balancer module (reference REF3020 40uA, divider 20+20kOhm 85uA)
// add info led, sviet 1s po starte a potom 0,1s po kazdom prebudeni (cca kazdu 1s)
// version 2 from 9.3.2019

#include "ADConversion.h"
#include <avr\sleep.h>
#include <ATTinyCore.h>
#include <Arduino.h>


//balancer ovladany pinom 7  (fyzicke piny na puzdre DIP8)
#define BALANACER_OUTPUT PB2

//signalizacia podpatia pint 6
#define UNDERVOLTAGE_OUTPUT PB1

//signalizna led pin 2
#define LED_OUTPUT PB3

//set balance voltage mV - balancovacie napatie
const uint16_t BALANCE_VOLTAGE = 3450;

//set undervoltage mV - podpatie
const uint16_t UNDER_VOLTAGE = 2900;

//napatova hysterezia 20mV
const uint8_t BALANCE_HYSTEREZIS = 20;

//tu sa da nastavit korekcia referencie
int8_t ref_correction = 0;


// vysledok posledneho merania
uint16_t actualVoltage = 0;
//vysledok n-1 merani
uint16_t previousVoltage = 0;

unsigned long ledOnAt = 0;

bool startup = true;
// Variables for the Sleep/power down modes:
volatile boolean f_wdt = 1;
// class for ADC
ADConversion adc;

// the setup function runs once when you press reset or power the board
void setup() {
  
  //default nastav vystupy na low
  pinMode(BALANACER_OUTPUT, OUTPUT);
  digitalWrite(BALANACER_OUTPUT, LOW);

  pinMode(UNDERVOLTAGE_OUTPUT, OUTPUT);
  digitalWrite(UNDERVOLTAGE_OUTPUT, LOW);
  //LED-Ka rozsviet pri inicializacii a zhasni az po 1s
  pinMode(LED_OUTPUT, OUTPUT);
  digitalWrite(LED_OUTPUT, HIGH);

  startup = true;
  //set watch dog to 8(4s), 7(2s), 6(1s) - prebudenie kazdu sekundu
  setup_watchdog(6); 
}

// Watchdog Interrupt Service / is executed when watchdog timed out
ISR(WDT_vect) {
	f_wdt = 1;  // set global flag
}

// the loop function runs over and over again forever
void loop() 
{
	if (f_wdt == 1)
	{  // wait for timed out watchdog / flag is set when a watchdog timeout occurs
		f_wdt = 0;       // reset flag, othervise processor will reset

		//start ADC - start prevodnika
		adc.start();
		// ignoruje prve prerusenie pri starte
		if (!startup)
		{
			//rozsviet LED - dufam ze to nebude ovplyvnovat AD vstup
			digitalWrite(LED_OUTPUT, HIGH);
			ledOnAt = millis();
		}
	}
	unsigned long actualms = millis();
	//zhasni led po cca 0,1s pokial ide o bezny beh a 1s ak ide o start
	if (digitalRead(LED_OUTPUT) && 
	((startup && (actualms - ledOnAt >= 1000)) || ((!startup) && (actualms - ledOnAt >= 100))))
	{
		digitalWrite(LED_OUTPUT, LOW);
		startup = false;
	}

	if (!adc.isConverting())
	{
		//adc is finished or stopped
		int adcVal = adc.getMeasuredValue();
		if (adcVal >= 0)
		{
			//finished measurement
			//pouzivam externu refernciu 2048mV, delic 1:1
			//result = adc_val * Uref/1024*2= result * 2048L*2/1024
			uint16_t adcVoltage = adcVal * (2048L + ref_correction) / 512L; 
			//robim exponencialny rolling priemer za posledne 3 hodnoty
			//inicializacia ked nie je historia
			if (previousVoltage == 0)
				previousVoltage = adcVoltage;

			actualVoltage = (adcVoltage + previousVoltage * 2) / 3;
			previousVoltage = actualVoltage; //odloz aktualnu hodnotu do historie
			

			//riadene balanceru
			bool balancer_on = ((actualVoltage > BALANCE_VOLTAGE) ||
				(digitalRead((BALANACER_OUTPUT) == HIGH) && (actualVoltage + BALANCE_HYSTEREZIS) >= BALANCE_VOLTAGE));
			//riadenie podpatia
			bool undervoltage_on = ((actualVoltage < UNDER_VOLTAGE) || 
				(digitalRead((UNDERVOLTAGE_OUTPUT) == HIGH) && (actualVoltage - BALANCE_HYSTEREZIS) <= UNDER_VOLTAGE));

			setBalancerUndervoltage(balancer_on, undervoltage_on);
		}
		//else
		//{
		//	//adc stopped
		//}
	}
	//pokial ADC konvertuje alebo LED svieti nedovol zaspat
	if ((!adc.isConverting()) && (!digitalRead(LED_OUTPUT)))
	{
		//sleep and wait for Watchdog or serial interrupt
		system_sleep();  // Set the unit to sleep
	}
} //end of loop


// 0=16ms, 1=32ms,2=64ms,3=128ms,4=250ms,5=500ms
// 6=1 sec,7=2 sec, 8=4 sec, 9= 8sec
void setup_watchdog(int ii) {

	byte bb;
	int ww;
	if (ii > 9) ii = 9;
	bb = ii & 7;
	if (ii > 7) bb |= (1 << 5);
	bb |= (1 << WDCE);
	ww = bb;
	//always clear the WDRF flag and the WDE control bit in the initialization routine
	MCUSR &= ~(1 << WDRF);
	// start timed sequence
	WDTCR |= (1 << WDCE) | (1 << WDE);
	// set new watchdog timeout value
	WDTCR = bb;
	WDTCR |= _BV(WDIE);
}

// set system into the sleep state 
// system wakes up when watchdog is timed out
void system_sleep() {
	//set power down mode
	//disable ADC
	ADCSRA &= ~(1 << ADEN);
	// set sleep mode
	set_sleep_mode(SLEEP_MODE_PWR_DOWN); 
	//  enable sleep 
	sleep_enable();
	// start sleep
	sleep_mode();                        // System actually sleeps here
	// disable sleep
	sleep_disable();                     // System continues execution here when watchdog timed out 
	//switch on ADC - not required it would be done in adc.Start
}

void setBalancerUndervoltage(bool balancer, bool undervoltage)
{
	if (balancer)
		digitalWrite(BALANACER_OUTPUT, HIGH);
	else
		digitalWrite(BALANACER_OUTPUT, LOW);

	if (undervoltage)
		digitalWrite(UNDERVOLTAGE_OUTPUT, HIGH);
	else
		digitalWrite(UNDERVOLTAGE_OUTPUT, LOW);
}

/* end of file */
